@@ -37,7 +37,7 @@
 #include "gatchat.h"
 #include "gatresult.h"
 
-#include "at.h"
+#include "atmodem.h"
 
 /* Amount of ms we wait between CLCC calls */
 #define POLL_CLCC_INTERVAL 500
@@ -89,51 +89,6 @@ static int class_to_call_type(int cls)
 	}
 }
 
-static unsigned int alloc_next_id(struct voicecall_data *d)
-{
-	unsigned int i;
-
-	for (i = 1; i < sizeof(d->id_list) * 8; i++) {
-		if (d->id_list & (0x1 << i))
-			continue;
-
-		d->id_list |= (0x1 << i);
-		return i;
-	}
-
-	return 0;
-}
-
-static void release_id(struct voicecall_data *d, unsigned int id)
-{
-	d->id_list &= ~(0x1 << id);
-}
-
-static gint call_compare_by_status(gconstpointer a, gconstpointer b)
-{
-	const struct ofono_call *call = a;
-	int status = GPOINTER_TO_INT(b);
-
-	if (status != call->status)
-		return 1;
-
-	return 0;
-}
-
-static gint call_compare(gconstpointer a, gconstpointer b)
-{
-	const struct ofono_call *ca = a;
-	const struct ofono_call *cb = b;
-
-	if (ca->id < cb->id)
-		return -1;
-
-	if (ca->id > cb->id)
-		return 1;
-
-	return 0;
-}
-
 static struct ofono_call *create_call(struct voicecall_data *d, int type,
 					int direction, int status,
 					const char *num, int num_type, int clip)
@@ -146,7 +101,7 @@ static struct ofono_call *create_call(struct voicecall_data *d, int type,
 	if (!call)
 		return NULL;
 
-	call->id = alloc_next_id(d);
+	call->id = at_util_alloc_next_id(&d->id_list);
 	call->type = type;
 	call->direction = direction;
 	call->status = status;
@@ -159,7 +114,7 @@ static struct ofono_call *create_call(struct voicecall_data *d, int type,
 
 	call->clip_validity = clip;
 
-	d->calls = g_slist_insert_sorted(d->calls, call, call_compare);
+	d->calls = g_slist_insert_sorted(d->calls, call, at_util_call_compare);
 
 	return call;
 }
@@ -213,7 +168,7 @@ static GSList *parse_clcc(GAtResult *result)
 		else
 			call->clip_validity = 2;
 
-		l = g_slist_insert_sorted(l, call, call_compare);
+		l = g_slist_insert_sorted(l, call, at_util_call_compare);
 	}
 
 	return l;
@@ -260,7 +215,7 @@ static void clcc_poll_cb(gboolean ok, GAtResult *result, gpointer user_data)
 				ofono_voicecall_disconnected(vc, oc->id,
 								reason, NULL);
 
-			release_id(vd, oc->id);
+			at_util_release_id(&vd->id_list, oc->id);
 
 			o = o->next;
 		} else if (nc && (!oc || (nc->id < oc->id))) {
@@ -467,10 +422,7 @@ error:
 	if (cbd)
 		g_free(cbd);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, data);
 }
 
 static void at_template(const char *cmd, struct ofono_voicecall *vc,
@@ -496,10 +448,7 @@ error:
 	if (req)
 		g_free(req);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, data);
 }
 
 static void at_answer(struct ofono_voicecall *vc,
@@ -536,16 +485,14 @@ static void clcc_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	calls = parse_clcc(result);
 
 	if (calls == NULL) {
-		DECLARE_FAILURE(e);
-		cb(&e, 0, NULL, cbd->data);
+		CALLBACK_WITH_FAILURE(cb, 0, NULL, cbd->data);
 		goto out;
 	}
 
 	list = g_try_new0(struct ofono_call, g_slist_length(calls));
 
 	if (!list) {
-		DECLARE_FAILURE(e);
-		cb(&e, 0, NULL, cbd->data);
+		CALLBACK_WITH_FAILURE(cb, 0, NULL, cbd->data);
 		goto out;
 	}
 
@@ -578,11 +525,7 @@ error:
 	if (cbd)
 		g_free(cbd);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, 0, NULL, data);
-	}
-
+	CALLBACK_WITH_FAILURE(cb, 0, NULL, data);
 }
 
 static void at_hold_all_active(struct ofono_voicecall *vc,
@@ -637,10 +580,7 @@ error:
 	if (req)
 		g_free(req);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, data);
 }
 
 static void at_private_chat(struct ofono_voicecall *vc, int id,
@@ -731,10 +671,7 @@ error:
 	if (cbd)
 		g_free(cbd);
 
-	{
-		DECLARE_FAILURE(error);
-		cb(&error, data);
-	}
+	CALLBACK_WITH_FAILURE(cb, data);
 }
 
 static void ring_notify(GAtResult *result, gpointer user_data)
@@ -745,9 +682,14 @@ static void ring_notify(GAtResult *result, gpointer user_data)
 
 	dump_response("ring_notify", TRUE, result);
 
+	/* See comment in CRING */
+	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(5),
+				at_util_call_compare_by_status))
+		return;
+
 	/* RING can repeat, ignore if we already have an incoming call */
 	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(4),
-				call_compare_by_status))
+				at_util_call_compare_by_status))
 		return;
 
 	/* Generate an incoming call of unknown type */
@@ -773,9 +715,20 @@ static void cring_notify(GAtResult *result, gpointer user_data)
 
 	dump_response("cring_notify", TRUE, result);
 
+	/* Handle the following situation:
+	 * Active Call + Waiting Call.  Active Call is Released.  The Waiting
+	 * call becomes Incoming and RING/CRING indications are signaled.
+	 * Sometimes these arrive before we managed to poll CLCC to find about
+	 * the stage change.  If this happens, simply ignore the RING/CRING
+	 * when a waiting call exists (cannot have waiting + incoming in GSM)
+	 */
+	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(5),
+				at_util_call_compare_by_status))
+		return;
+
 	/* CRING can repeat, ignore if we already have an incoming call */
 	if (g_slist_find_custom(vd->calls, GINT_TO_POINTER(4),
-				call_compare_by_status))
+				at_util_call_compare_by_status))
 		return;
 
 	g_at_result_iter_init(&iter, result);
@@ -820,7 +773,7 @@ static void clip_notify(GAtResult *result, gpointer user_data)
 	dump_response("clip_notify", TRUE, result);
 
 	l = g_slist_find_custom(vd->calls, GINT_TO_POINTER(4),
-				call_compare_by_status);
+				at_util_call_compare_by_status);
 
 	if (l == NULL) {
 		ofono_error("CLIP for unknown call");
@@ -997,7 +950,6 @@ static int at_voicecall_probe(struct ofono_voicecall *vc, unsigned int vendor,
 	g_at_chat_send(chat, "AT+CRC=1", NULL, NULL, NULL, NULL);
 	g_at_chat_send(chat, "AT+CLIP=1", NULL, NULL, NULL, NULL);
 	g_at_chat_send(chat, "AT+COLP=1", NULL, NULL, NULL, NULL);
-	g_at_chat_send(chat, "AT+CSSN=1,1", NULL, NULL, NULL, NULL);
 	g_at_chat_send(chat, "AT+CCWA=1", NULL,
 				at_voicecall_initialized, vc, NULL);
 	return 0;
