@@ -23,16 +23,36 @@
 #include <config.h>
 #endif
 
+#define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <dirent.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+
 #include <glib.h>
 
 #include "util.h"
+#include "storage.h"
 #include "smsutil.h"
 
 #define uninitialized_var(x) x = x
 
+#define SMS_BACKUP_MODE 0600
+#define SMS_BACKUP_PATH STORAGEDIR "/%s/sms"
+#define SMS_BACKUP_PATH_DIR SMS_BACKUP_PATH "/%s-%i-%i"
+#define SMS_BACKUP_PATH_FILE SMS_BACKUP_PATH_DIR "/%03i"
+
+#define SMS_ADDR_FMT "%24[0-9A-F]"
+
+static GSList *sms_assembly_add_fragment_backup(struct sms_assembly *assembly,
+					const struct sms *sms, time_t ts,
+					const struct sms_address *addr,
+					guint16 ref, guint8 max, guint8 seq,
+					gboolean backup);
+
 void extract_bcd_number(const unsigned char *buf, int len, char *out)
 {
 	static const char digit_lut[] = "0123456789*#abc\0";
@@ -110,6 +130,17 @@ static inline int to_semi_oct(char in)
 	return digit;
 }
 
+void encode_bcd_number(const char *number, unsigned char *out)
+{
+	while (number[0] != '\0' && number[1] != '\0') {
+		*out = to_semi_oct(*number++);
+		*out++ |= to_semi_oct(*number++) << 4;
+	}
+
+	if (*number)
+		*out = to_semi_oct(*number) | 0xf0;
+}
+
 /* Returns whether the DCS could be parsed successfully, e.g. no reserved
  * values were used
  */
@@ -446,8 +477,8 @@ static gboolean encode_validity_period(const struct sms_validity_period *vp,
 	return FALSE;
 }
 
-static gboolean encode_address(const struct sms_address *in, gboolean sc,
-				unsigned char *pdu, int *offset)
+gboolean sms_encode_address_field(const struct sms_address *in, gboolean sc,
+					unsigned char *pdu, int *offset)
 {
 	size_t len = strlen(in->address);
 	unsigned char addr_len = 0;
@@ -533,9 +564,9 @@ out:
 	return TRUE;
 }
 
-static gboolean decode_address(const unsigned char *pdu, int len,
-				int *offset, gboolean sc,
-				struct sms_address *out)
+gboolean sms_decode_address_field(const unsigned char *pdu, int len,
+					int *offset, gboolean sc,
+					struct sms_address *out)
 {
 	unsigned char addr_len;
 	unsigned char addr_type;
@@ -564,7 +595,7 @@ static gboolean decode_address(const unsigned char *pdu, int len,
 	out->numbering_plan = bit_field(addr_type, 0, 4);
 
 	if (out->number_type != SMS_NUMBER_TYPE_ALPHANUMERIC) {
-		extract_bcd_number(pdu+*offset, byte_len, out->address);
+		extract_bcd_number(pdu + *offset, byte_len, out->address);
 		*offset += byte_len;
 	} else {
 		int chars;
@@ -635,7 +666,7 @@ static gboolean encode_deliver(const struct sms_deliver *in, unsigned char *pdu,
 
 	set_octet(pdu, offset, oct);
 
-	if (encode_address(&in->oaddr, FALSE, pdu, offset) == FALSE)
+	if (sms_encode_address_field(&in->oaddr, FALSE, pdu, offset) == FALSE)
 		return FALSE;
 
 	set_octet(pdu, offset, in->pid);
@@ -672,7 +703,8 @@ static gboolean decode_deliver(const unsigned char *pdu, int len,
 	out->deliver.udhi = is_bit_set(octet, 6);
 	out->deliver.rp = is_bit_set(octet, 7);
 
-	if (!decode_address(pdu, len, &offset, FALSE, &out->deliver.oaddr))
+	if (!sms_decode_address_field(pdu, len, &offset,
+					FALSE, &out->deliver.oaddr))
 		return FALSE;
 
 	if (!next_octet(pdu, len, &offset, &out->deliver.pid))
@@ -773,7 +805,7 @@ static gboolean decode_submit_report(const unsigned char *pdu, int len,
 	int offset = 0;
 	unsigned char octet;
 	gboolean udhi;
-	guint8 fcs;
+	guint8 uninitialized_var(fcs);
 	guint8 pi;
 	struct sms_scts *scts;
 	guint8 pid = 0;
@@ -883,7 +915,7 @@ static gboolean encode_status_report(const struct sms_status_report *in,
 
 	set_octet(pdu, offset, in->mr);
 
-	if (!encode_address(&in->raddr, FALSE, pdu, offset))
+	if (!sms_encode_address_field(&in->raddr, FALSE, pdu, offset))
 		return FALSE;
 
 	if (!encode_scts(&in->scts, pdu, offset))
@@ -935,8 +967,8 @@ static gboolean decode_status_report(const unsigned char *pdu, int len,
 	if (!next_octet(pdu, len, &offset, &out->status_report.mr))
 		return FALSE;
 
-	if (!decode_address(pdu, len, &offset, FALSE,
-				&out->status_report.raddr))
+	if (!sms_decode_address_field(pdu, len, &offset, FALSE,
+					&out->status_report.raddr))
 		return FALSE;
 
 	if (!decode_scts(pdu, len, &offset, &out->status_report.scts))
@@ -1064,7 +1096,7 @@ static gboolean decode_deliver_report(const unsigned char *pdu, int len,
 	int offset = 0;
 	unsigned char octet;
 	gboolean udhi;
-	guint8 fcs;
+	guint8 uninitialized_var(fcs);
 	guint8 pi;
 	guint8 pid = 0;
 	guint8 dcs = 0;
@@ -1173,7 +1205,7 @@ static gboolean encode_submit(const struct sms_submit *in,
 
 	set_octet(pdu, offset, in->mr);
 
-	if (encode_address(&in->daddr, FALSE, pdu, offset) == FALSE)
+	if (sms_encode_address_field(&in->daddr, FALSE, pdu, offset) == FALSE)
 		return FALSE;
 
 	set_octet(pdu, offset, in->pid);
@@ -1215,7 +1247,8 @@ static gboolean decode_submit(const unsigned char *pdu, int len,
 	if (!next_octet(pdu, len, &offset, &out->submit.mr))
 		return FALSE;
 
-	if (!decode_address(pdu, len, &offset, FALSE, &out->submit.daddr))
+	if (!sms_decode_address_field(pdu, len, &offset,
+					FALSE, &out->submit.daddr))
 		return FALSE;
 
 	if (!next_octet(pdu, len, &offset, &out->submit.pid))
@@ -1265,7 +1298,7 @@ static gboolean encode_command(const struct sms_command *in,
 
 	set_octet(pdu, offset, in->mn);
 
-	if (!encode_address(&in->daddr, FALSE, pdu, offset))
+	if (!sms_encode_address_field(&in->daddr, FALSE, pdu, offset))
 		return FALSE;
 
 	set_octet(pdu, offset, in->cdl);
@@ -1305,7 +1338,8 @@ static gboolean decode_command(const unsigned char *pdu, int len,
 	if (!next_octet(pdu, len, &offset, &out->command.mn))
 		return FALSE;
 
-	if (!decode_address(pdu, len, &offset, FALSE, &out->command.daddr))
+	if (!sms_decode_address_field(pdu, len, &offset,
+					FALSE, &out->command.daddr))
 		return FALSE;
 
 	if (!next_octet(pdu, len, &offset, &out->command.cdl))
@@ -1328,7 +1362,7 @@ gboolean sms_encode(const struct sms *in, int *len, int *tpdu_len,
 
 	if (in->type == SMS_TYPE_DELIVER || in->type == SMS_TYPE_SUBMIT ||
 			in->type == SMS_TYPE_COMMAND)
-		if (!encode_address(&in->sc_addr, TRUE, pdu, &offset))
+		if (!sms_encode_address_field(&in->sc_addr, TRUE, pdu, &offset))
 			return FALSE;
 
 	tpdu_start = offset;
@@ -1396,7 +1430,8 @@ gboolean sms_decode(const unsigned char *pdu, int len, gboolean outgoing,
 		return FALSE;
 
 	if (tpdu_len < len) {
-		if (!decode_address(pdu, len, &offset, TRUE, &out->sc_addr))
+		if (!sms_decode_address_field(pdu, len, &offset,
+						TRUE, &out->sc_addr))
 			return FALSE;
 	}
 
@@ -1921,6 +1956,59 @@ gboolean sms_extract_concatenation(const struct sms *sms, guint16 *ref_num,
 	return TRUE;
 }
 
+gboolean sms_extract_language_variant(const struct sms *sms, guint8 *locking,
+					guint8 *single)
+{
+	struct sms_udh_iter iter;
+	enum sms_iei iei;
+	guint8 variant;
+
+	/* We must ignore the entire user_data header here:
+	 * If the length of the User Data Header is such that there
+	 * are too few or too many octets in the final Information
+	 * Element then the whole User Data Header shall be ignored.
+	 */
+	if (!sms_udh_iter_init(sms, &iter))
+		return FALSE;
+
+	/* According to the specification, we have to use the last
+	 * useable header:
+	 * In the event that IEs determined as not repeatable are
+	 * duplicated, the last occurrence of the IE shall be used.
+	 * In the event that two or more IEs occur which have mutually
+	 * exclusive meanings (e.g. an 8bit port address and a 16bit
+	 * port address), then the last occurring IE shall be used.
+	 */
+	while ((iei = sms_udh_iter_get_ie_type(&iter)) !=
+			SMS_IEI_INVALID) {
+		switch (iei) {
+		case SMS_IEI_NATIONAL_LANGUAGE_SINGLE_SHIFT:
+			if (sms_udh_iter_get_ie_length(&iter) != 1)
+				break;
+
+			sms_udh_iter_get_ie_data(&iter, &variant);
+			if (single)
+				*single = variant;
+			break;
+
+		case SMS_IEI_NATIONAL_LANGUAGE_LOCKING_SHIFT:
+			if (sms_udh_iter_get_ie_length(&iter) != 1)
+				break;
+
+			sms_udh_iter_get_ie_data(&iter, &variant);
+			if (locking)
+				*locking = variant;
+			break;
+		default:
+			break;
+		}
+
+		sms_udh_iter_next(&iter);
+	}
+
+	return TRUE;
+}
+
 /*!
  * Decodes a list of SMSes that contain a datagram.  The list must be
  * sorted in order of the sequence number.  This function assumes that
@@ -2045,21 +2133,40 @@ char *sms_decode_text(GSList *sms_list)
 
 		udl_in_bytes = sms_udl_in_bytes(udl, dcs);
 
+		if (udl_in_bytes == taken)
+			continue;
+
 		if (charset == SMS_CHARSET_7BIT) {
 			unsigned char buf[160];
 			long written;
+			guint8 locking_shift = 0;
+			guint8 single_shift = 0;
 			int max_chars = sms_text_capacity_gsm(udl, taken);
 
-			unpack_7bit_own_buf(ud + taken, udl_in_bytes - taken,
+			if (unpack_7bit_own_buf(ud + taken,
+						udl_in_bytes - taken,
 						taken, FALSE, max_chars,
-						&written, 0, buf);
+						&written, 0, buf) == NULL)
+				continue;
 
 			/* Take care of improperly split fragments */
 			if (buf[written-1] == 0x1b)
 				written = written - 1;
 
-			converted = convert_gsm_to_utf8(buf, written,
-							NULL, NULL, 0);
+			sms_extract_language_variant(sms, &locking_shift, &single_shift);
+
+			/* If language is not defined in 3GPP TS 23.038,
+	 		 * implementations are instructed to ignore it' */
+			if (locking_shift >= GSM_DIALECT_INVALID)
+				locking_shift = GSM_DIALECT_DEFAULT;
+
+			if (single_shift >= GSM_DIALECT_INVALID)
+				single_shift = GSM_DIALECT_DEFAULT;
+
+			converted = convert_gsm_to_utf8_with_lang(buf, written,
+								NULL, NULL, 0,
+								locking_shift,
+								single_shift);
 		} else {
 			const gchar *from = (const gchar *)(ud + taken);
 			/* According to the spec: A UCS2 character shall not be
@@ -2083,17 +2190,217 @@ char *sms_decode_text(GSList *sms_list)
 
 	utf8 = g_string_free(str, FALSE);
 
-	if (strlen(utf8) == 0) {
-		g_free(utf8);
-		return NULL;
-	}
-
 	return utf8;
 }
 
-struct sms_assembly *sms_assembly_new()
+static int sms_serialize(unsigned char *buf, const struct sms *sms)
 {
-	return g_new0(struct sms_assembly, 1);
+	int len, tpdu_len;
+
+	sms_encode(sms, &len, &tpdu_len, buf + 1);
+	buf[0] = tpdu_len;
+
+	return len + 1;
+}
+
+static gboolean sms_deserialize(const unsigned char *buf,
+		struct sms *sms, int len)
+{
+	if (len < 1)
+		return FALSE;
+
+	return sms_decode(buf + 1, len - 1, FALSE, buf[0], sms);
+}
+
+static gboolean sms_assembly_extract_address(const char *straddr,
+						struct sms_address *out)
+{
+	unsigned char pdu[12];
+	long len;
+	int offset = 0;
+
+	if (decode_hex_own_buf(straddr, -1, &len, 0, pdu) == NULL)
+		return FALSE;
+
+	return sms_decode_address_field(pdu, len, &offset, FALSE, out);
+}
+
+static gboolean sms_assembly_encode_address(const struct sms_address *in,
+						char *straddr)
+{
+	unsigned char pdu[12];
+	int offset = 0;
+
+	if (sms_encode_address_field(in, FALSE, pdu, &offset) == FALSE)
+		return FALSE;
+
+	if (encode_hex_own_buf(pdu, offset, 0, straddr) == NULL)
+		return FALSE;
+
+	straddr[offset * 2 + 1] = '\0';
+
+	return TRUE;
+}
+
+static void sms_assembly_load(struct sms_assembly *assembly,
+				const struct dirent *dir)
+{
+	struct sms_address addr;
+	char straddr[25];
+	guint16 ref;
+	guint8 max;
+	guint8 seq;
+	char *path;
+	int len;
+	struct stat segment_stat;
+	struct dirent **segments;
+	char *endp;
+	int r;
+	int i;
+	unsigned char buf[177];
+	struct sms segment;
+
+	if (dir->d_type != DT_DIR)
+		return;
+
+	/* Max of SMS address size is 12 bytes, hex encoded */
+	if (sscanf(dir->d_name, SMS_ADDR_FMT "-%hi-%hhi",
+				straddr, &ref, &max) < 3)
+		return;
+
+	if (sms_assembly_extract_address(straddr, &addr) == FALSE)
+		return;
+
+	path = g_strdup_printf(SMS_BACKUP_PATH "/%s",
+			assembly->imsi, dir->d_name);
+	len = scandir(path, &segments, NULL, versionsort);
+	g_free(path);
+
+	if (len < 0)
+		return;
+
+	for (i = 0; i < len; i++) {
+		if (segments[i]->d_type != DT_REG)
+			continue;
+
+		seq = strtol(segments[i]->d_name, &endp, 10);
+		if (*endp != '\0')
+			continue;
+
+		r = read_file(buf, sizeof(buf), SMS_BACKUP_PATH "/%s/%s",
+				assembly->imsi,
+				dir->d_name, segments[i]->d_name);
+		if (r < 0)
+			continue;
+
+		if (!sms_deserialize(buf, &segment, r))
+			continue;
+
+		path = g_strdup_printf(SMS_BACKUP_PATH "/%s/%s",
+				assembly->imsi,
+				dir->d_name, segments[i]->d_name);
+		r = stat(path, &segment_stat);
+		g_free(path);
+
+		if (r != 0)
+			continue;
+
+		/* Errors cannot occur here */
+		sms_assembly_add_fragment_backup(assembly, &segment,
+						segment_stat.st_mtime,
+						&addr, ref, max, seq, FALSE);
+	}
+
+	for (i = 0; i < len; i++)
+		free(segments[i]);
+
+	free(segments);
+}
+
+static gboolean sms_assembly_store(struct sms_assembly *assembly,
+				struct sms_assembly_node *node,
+				const struct sms *sms, guint8 seq)
+{
+	unsigned char buf[177];
+	int len;
+	char straddr[25];
+
+	if (!assembly->imsi)
+		return FALSE;
+
+	if (sms_assembly_encode_address(&node->addr, straddr) == FALSE)
+		return FALSE;
+
+	len = sms_serialize(buf, sms);
+
+	if (write_file(buf, len, SMS_BACKUP_MODE,
+				SMS_BACKUP_PATH_FILE, assembly->imsi, straddr,
+				node->ref, node->max_fragments, seq) != len)
+		return FALSE;
+
+	return TRUE;
+}
+
+static void sms_assembly_backup_free(struct sms_assembly *assembly,
+					struct sms_assembly_node *node)
+{
+	char *path;
+	int seq;
+	char straddr[25];
+
+	if (!assembly->imsi)
+		return;
+
+	if (sms_assembly_encode_address(&node->addr, straddr) == FALSE)
+		return;
+
+	for (seq = 0; seq < node->max_fragments; seq++) {
+		int offset = seq / 32;
+		int bit = 1 << (seq % 32);
+
+		if (node->bitmap[offset] & bit) {
+			path = g_strdup_printf(SMS_BACKUP_PATH_FILE,
+					assembly->imsi, straddr,
+					node->ref, node->max_fragments, seq);
+			unlink(path);
+			g_free(path);
+		}
+	}
+
+	path = g_strdup_printf(SMS_BACKUP_PATH_DIR, assembly->imsi, straddr,
+				node->ref, node->max_fragments);
+	rmdir(path);
+	g_free(path);
+}
+
+struct sms_assembly *sms_assembly_new(const char *imsi)
+{
+	struct sms_assembly *ret = g_new0(struct sms_assembly, 1);
+	char *path;
+	struct dirent **entries;
+	int len;
+
+	if (imsi) {
+		ret->imsi = imsi;
+
+		/* Restore state from backup */
+
+		path = g_strdup_printf(SMS_BACKUP_PATH, imsi);
+		len = scandir(path, &entries, NULL, alphasort);
+		g_free(path);
+
+		if (len < 0)
+			return ret;
+
+		while (len--) {
+			sms_assembly_load(ret, entries[len]);
+			free(entries[len]);
+		}
+
+		free(entries);
+	}
+
+	return ret;
 }
 
 void sms_assembly_free(struct sms_assembly *assembly)
@@ -2117,7 +2424,17 @@ GSList *sms_assembly_add_fragment(struct sms_assembly *assembly,
 					const struct sms_address *addr,
 					guint16 ref, guint8 max, guint8 seq)
 {
-	int offset = seq / 8;
+	return sms_assembly_add_fragment_backup(assembly, sms,
+						ts, addr, ref, max, seq, TRUE);
+}
+
+static GSList *sms_assembly_add_fragment_backup(struct sms_assembly *assembly,
+					const struct sms *sms, time_t ts,
+					const struct sms_address *addr,
+					guint16 ref, guint8 max, guint8 seq,
+					gboolean backup)
+{
+	int offset = seq / 32;
 	int bit = 1 << (seq % 32);
 	GSList *l;
 	GSList *prev;
@@ -2190,11 +2507,17 @@ out:
 	node->bitmap[offset] |= bit;
 	node->num_fragments += 1;
 
-	if (node->num_fragments < node->max_fragments)
+	if (node->num_fragments < node->max_fragments) {
+		if (backup)
+			sms_assembly_store(assembly, node, sms, seq);
+
 		return NULL;
+	}
 
 	completed = node->fragment_list;
 
+	sms_assembly_backup_free(assembly, node);
+
 	if (prev)
 		prev->next = l->next;
 	else
@@ -2228,6 +2551,8 @@ void sms_assembly_expire(struct sms_assembly *assembly, time_t before)
 			continue;
 		}
 
+		sms_assembly_backup_free(assembly, node);
+
 		g_slist_foreach(node->fragment_list, (GFunc)g_free, 0);
 		g_slist_free(node->fragment_list);
 		g_free(node);
@@ -2586,7 +2911,7 @@ gboolean cbs_extract_app_port(const struct cbs *cbs, int *dst, int *src,
 	return extract_app_port_common(&iter, dst, src, is_8bit);
 }
 
-static gboolean iso639_2_from_language(enum cbs_language lang, char *iso639)
+gboolean iso639_2_from_language(enum cbs_language lang, char *iso639)
 {
 	switch (lang) {
 	case CBS_LANGUAGE_GERMAN:
@@ -2972,7 +3297,7 @@ static void cbs_assembly_expire(struct cbs_assembly *assembly,
 	}
 }
 
-void cbs_assembly_location_changed(struct cbs_assembly *assembly,
+void cbs_assembly_location_changed(struct cbs_assembly *assembly, gboolean plmn,
 					gboolean lac, gboolean ci)
 {
 	/* Location Area wide (in GSM) (which means that a CBS message with the
@@ -2989,6 +3314,15 @@ void cbs_assembly_location_changed(struct cbs_assembly *assembly,
 	 * one cell only.
 	 */
 
+	if (plmn) {
+		lac = TRUE;
+		g_slist_free(assembly->recv_plmn);
+		assembly->recv_plmn = NULL;
+
+		cbs_assembly_expire(assembly, cbs_compare_node_by_gs,
+				GUINT_TO_POINTER(CBS_GEO_SCOPE_PLMN));
+	}
+
 	if (lac) {
 		/* If LAC changed, then cell id has changed */
 		ci = TRUE;
@@ -3111,3 +3445,224 @@ out:
 
 	return completed;
 }
+
+static inline int skip_to_next_field(const char *str, int pos, int len)
+{
+	if (pos < len && str[pos] == ',')
+		pos += 1;
+
+	while (pos < len && str[pos] == ' ')
+		pos += 1;
+
+	return pos;
+}
+
+static gboolean next_range(const char *str, int *offset, gint *min, gint *max)
+{
+	int pos;
+	int end;
+	int len;
+	int low = 0;
+	int high = 0;
+
+	len = strlen(str);
+
+	pos = *offset;
+
+	while (pos < len && str[pos] == ' ')
+		pos += 1;
+
+	end = pos;
+
+	while (str[end] >= '0' && str[end] <= '9') {
+		low = low * 10 + (int)(str[end] - '0');
+		end += 1;
+	}
+
+	if (pos == end)
+		return FALSE;
+
+	if (str[end] != '-') {
+		high = low;
+		goto out;
+	}
+
+	pos = end = end + 1;
+
+	while (str[end] >= '0' && str[end] <= '9') {
+		high = high * 10 + (int)(str[end] - '0');
+		end += 1;
+	}
+
+	if (pos == end)
+		return FALSE;
+
+out:
+	*offset = skip_to_next_field(str, end, len);
+
+	if (min)
+		*min = low;
+
+	if (max)
+		*max = high;
+
+	return TRUE;
+}
+
+static GSList *cbs_optimize_ranges(GSList *ranges)
+{
+	struct cbs_topic_range *range;
+	unsigned char bitmap[125];
+	GSList *l;
+	unsigned short i;
+	GSList *ret = NULL;
+
+	memset(bitmap, 0, sizeof(bitmap));
+
+	for (l = ranges; l; l = l->next) {
+		range = l->data;
+
+		for (i = range->min; i <= range->max; i++) {
+			int byte_offset = i / 8;
+			int bit = i % 8;
+
+			bitmap[byte_offset] |= 1 << bit;
+		}
+	}
+
+	range = NULL;
+
+	for (i = 0; i <= 999; i++) {
+		int byte_offset = i / 8;
+		int bit = i % 8;
+
+		if (is_bit_set(bitmap[byte_offset], bit) == FALSE) {
+			if (range) {
+				ret = g_slist_prepend(ret, range);
+				range = NULL;
+			}
+
+			continue;
+		}
+
+		if (range) {
+			range->max = i;
+			continue;
+		}
+
+		range = g_new0(struct cbs_topic_range, 1);
+		range->min = i;
+		range->max = i;
+	}
+
+	if (range != NULL)
+		ret = g_slist_prepend(ret, range);
+
+	ret = g_slist_reverse(ret);
+
+	return ret;
+}
+
+GSList *cbs_extract_topic_ranges(const char *ranges)
+{
+	int min;
+	int max;
+	int offset = 0;
+	GSList *ret = NULL;
+	GSList *tmp;
+
+	while (next_range(ranges, &offset, &min, &max) == TRUE) {
+		if (min < 0 || min > 999)
+			return NULL;
+
+		if (max < 0 || max > 999)
+			return NULL;
+
+		if (max < min)
+			return NULL;
+	}
+
+	if (ranges[offset] != '\0')
+		return NULL;
+
+	offset = 0;
+
+	while (next_range(ranges, &offset, &min, &max) == TRUE) {
+		struct cbs_topic_range *range = g_new0(struct cbs_topic_range, 1);
+
+		range->min = min;
+		range->max = max;
+
+		ret = g_slist_prepend(ret, range);
+	}
+
+	tmp = cbs_optimize_ranges(ret);
+	g_slist_foreach(ret, (GFunc)g_free, NULL);
+	g_slist_free(ret);
+
+	return tmp;
+}
+
+static inline int element_length(unsigned short element)
+{
+	if (element <= 9)
+		return 1;
+
+	if (element <= 99)
+		return 2;
+
+	if (element <= 999)
+		return 3;
+
+	if (element <= 9999)
+		return 4;
+
+	return 5;
+}
+
+static inline int range_length(struct cbs_topic_range *range)
+{
+	if (range->min == range->max)
+		return element_length(range->min);
+
+	return element_length(range->min) + element_length(range->max) + 1;
+}
+
+char *cbs_topic_ranges_to_string(GSList *ranges)
+{
+	int len = 0;
+	int nelem = 0;
+	struct cbs_topic_range *range;
+	GSList *l;
+	char *ret;
+
+	if (ranges == NULL)
+		return g_new0(char, 1);
+
+	for (l = ranges; l; l = l->next) {
+		range = l->data;
+
+		len += range_length(range);
+		nelem += 1;
+	}
+
+	/* Space for ranges, commas and terminator null */
+	ret = g_new(char, len + nelem);
+
+	len = 0;
+
+	for (l = ranges; l; l = l->next) {
+		range = l->data;
+
+		if (range->min != range->max)
+			len += sprintf(ret + len, "%hu-%hu",
+					range->min, range->max);
+		else
+			len += sprintf(ret + len, "%hu", range->min);
+
+		if (l->next != NULL)
+			ret[len++] = ',';
+	}
+
+	return ret;
+}
